# encoding:UTF-8
import ctypes
from ctypes import *
from datetime import datetime, timedelta
from decimal import Decimal

import pytz

from taos.constants import FieldType
from taos.error import DatabaseError

_priv_tz = None
_utc_tz = pytz.timezone("UTC")
try:
    _datetime_epoch = datetime.fromtimestamp(0)
except OSError:
    _datetime_epoch = datetime.fromtimestamp(86400) - timedelta(seconds=86400)
_utc_datetime_epoch = _utc_tz.localize(datetime.utcfromtimestamp(0))
_priv_datetime_epoch = None


def set_tz(tz):
    # type: (str) -> None
    global _priv_tz, _priv_datetime_epoch
    _priv_tz = tz
    _priv_datetime_epoch = _utc_datetime_epoch.astimezone(_priv_tz)


def get_tz():
    # type: (str) -> timezone
    return _priv_tz


def _convert_millisecond_to_datetime(milli):
    try:
        if _priv_tz is None:
            return _datetime_epoch + timedelta(seconds=milli / 1000.0)
        return _priv_datetime_epoch + timedelta(seconds=milli / 1000.0)
    except OverflowError:
        # catch OverflowError and pass
        print("WARN: datetime overflow!")
        pass


def _convert_microsecond_to_datetime(micro):
    try:
        if _priv_tz is None:
            return _datetime_epoch + timedelta(seconds=micro / 1000000.0)
        return _priv_datetime_epoch + timedelta(seconds=micro / 1000000.0)
    except OverflowError:
        # catch OverflowError and pass
        print("WARN: datetime overflow!")
        pass


def _convert_nanosecond_to_datetime(nanosec):
    return nanosec


def _crow_timestamp_to_python(data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C bool row to python row."""
    _timestamp_converter = _convert_millisecond_to_datetime
    if precision == FieldType.C_TIMESTAMP_MILLI:
        _timestamp_converter = _convert_millisecond_to_datetime
    elif precision == FieldType.C_TIMESTAMP_MICRO:
        _timestamp_converter = _convert_microsecond_to_datetime
    elif precision == FieldType.C_TIMESTAMP_NANO:
        _timestamp_converter = _convert_nanosecond_to_datetime
    else:
        raise DatabaseError("Unknown precision returned from database")

    return [
        None if is_null[i] else _timestamp_converter(ele)
        for i, ele in enumerate(ctypes.cast(data, ctypes.POINTER(ctypes.c_int64))[: abs(num_of_rows)])
    ]


def _crow_bool_to_python(data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C bool row to python row."""
    return [
        None if is_null[i] else bool(ele)
        for i, ele in enumerate(ctypes.cast(data, ctypes.POINTER(ctypes.c_byte))[: abs(num_of_rows)])
    ]


def _crow_tinyint_to_python(data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C tinyint row to python row."""
    return [
        None if is_null[i] else ele
        for i, ele in enumerate(ctypes.cast(data, ctypes.POINTER(ctypes.c_byte))[: abs(num_of_rows)])
    ]


def _crow_tinyint_unsigned_to_python(data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C unsigned tinyint row to python row."""
    return [
        None if is_null[i] else ele
        for i, ele in enumerate(ctypes.cast(data, ctypes.POINTER(ctypes.c_ubyte))[: abs(num_of_rows)])
    ]


def _crow_smallint_to_python(data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C smallint row to python row."""
    return [
        None if is_null[i] else ele
        for i, ele in enumerate(ctypes.cast(data, ctypes.POINTER(ctypes.c_short))[: abs(num_of_rows)])
    ]


def _crow_smallint_unsigned_to_python(
        data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN
):
    """Function to convert C unsigned smallint row to python row."""
    return [
        None if is_null[i] else ele
        for i, ele in enumerate(ctypes.cast(data, ctypes.POINTER(ctypes.c_ushort))[: abs(num_of_rows)])
    ]


def _crow_int_to_python(data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C int row to python row."""
    return [
        None if is_null[i] else ele
        for i, ele in enumerate(ctypes.cast(data, ctypes.POINTER(ctypes.c_int))[: abs(num_of_rows)])
    ]


def _crow_int_unsigned_to_python(data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C unsigned int row to python row."""
    return [
        None if is_null[i] else ele
        for i, ele in enumerate(ctypes.cast(data, ctypes.POINTER(ctypes.c_uint))[: abs(num_of_rows)])
    ]


def _crow_bigint_to_python(data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C bigint row to python row."""
    return [
        None if is_null[i] else ele
        for i, ele in enumerate(ctypes.cast(data, ctypes.POINTER(ctypes.c_int64))[: abs(num_of_rows)])
    ]


def _crow_bigint_unsigned_to_python(data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C unsigned bigint row to python row."""
    return [
        None if is_null[i] else ele
        for i, ele in enumerate(ctypes.cast(data, ctypes.POINTER(ctypes.c_uint64))[: abs(num_of_rows)])
    ]


def _crow_float_to_python(data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C float row to python row."""
    return [
        None if is_null[i] else ele
        for i, ele in enumerate(ctypes.cast(data, ctypes.POINTER(ctypes.c_float))[: abs(num_of_rows)])
    ]


def _crow_double_to_python(data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C double row to python row."""
    return [
        None if is_null[i] else ele
        for i, ele in enumerate(ctypes.cast(data, ctypes.POINTER(ctypes.c_double))[: abs(num_of_rows)])
    ]


def _crow_decimal_to_python(data, is_null, num_of_rows, nbytes=None, precision=0):
    """Function to convert C Decimal row to python row."""
    ptr = ctypes.cast(data, ctypes.POINTER(ctypes.c_char * 64))
    return [
        None if is_null[i] else Decimal(cast(ptr[i], c_char_p).value.decode("utf-8"))
        for i in range(abs(num_of_rows))
    ]


def _crow_binary_to_python(data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C binary row to python row."""
    assert nbytes is not None
    return [
        None if is_null[i] else ele.value.decode("utf-8")
        for i, ele in enumerate((ctypes.cast(data, ctypes.POINTER(ctypes.c_char * nbytes)))[: abs(num_of_rows)])
    ]


def _crow_varbinary_to_python(data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C binary row to python row."""
    assert nbytes is not None
    return [
        None if is_null[i] else ele.value
        for i, ele in enumerate((ctypes.cast(data, ctypes.POINTER(ctypes.c_char * nbytes)))[: abs(num_of_rows)])
    ]


def _crow_nchar_to_python(data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C nchar row to python row."""
    assert nbytes is not None
    res = []
    for i in range(abs(num_of_rows)):
        if is_null[i]:
            res.append(None)
        else:
            try:
                if num_of_rows >= 0:
                    tmpstr = ctypes.c_char_p(data)
                    res.append(tmpstr.value.decode("utf-8"))
                else:
                    res.append(
                        (
                            ctypes.cast(
                                data + nbytes * i,
                                ctypes.POINTER(ctypes.c_wchar * (nbytes // 4)),
                            )
                        )[0].value
                    )
            except ValueError:
                res.append(None)

    return res


def _crow_binary_to_python_block(data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C binary row to python row."""
    assert nbytes is not None
    res = []
    for i in range(abs(num_of_rows)):
        if is_null[i]:
            res.append(None)
        else:
            rbyte = ctypes.cast(data + nbytes * i, ctypes.POINTER(ctypes.c_uint16))[:1].pop()
            chars = ctypes.cast(c_char_p(data + nbytes * i + 2), ctypes.POINTER(c_char * rbyte))
            buffer = create_string_buffer(rbyte + 1)
            buffer[:rbyte] = chars[0][:rbyte]
            res.append(cast(buffer, c_char_p).value.decode("utf-8"))
    return res


def _crow_varbinary_to_python_block(data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C binary row to python row."""
    assert nbytes is not None
    res = []
    for i in range(abs(num_of_rows)):
        if is_null[i]:
            res.append(None)
        else:
            rbyte = ctypes.cast(data + nbytes * i, ctypes.POINTER(ctypes.c_uint16))[:1].pop()
            chars = ctypes.cast(c_char_p(data + nbytes * i + 2), ctypes.POINTER(c_char * rbyte))
            buffer = create_string_buffer(rbyte + 1)
            buffer[:rbyte] = chars[0][:rbyte]
            res.append(cast(buffer, c_char_p).value)
    return res


def _crow_nchar_to_python_block(data, is_null, num_of_rows, nbytes=None, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C nchar row to python row."""
    assert nbytes is not None
    res = []
    for i in range(abs(num_of_rows)):
        if is_null[i]:
            res.append(None)
        else:
            rbyte = ctypes.cast(data + nbytes * i, ctypes.POINTER(ctypes.c_uint16))[:1].pop()
            chars = ctypes.cast(c_char_p(data + nbytes * i + 2), ctypes.POINTER(c_char * rbyte))
            buffer = create_string_buffer(rbyte + 1)
            buffer[:rbyte] = chars[0][:rbyte]
            res.append(cast(buffer, c_char_p).value.decode("utf-8"))
    return res


def convert_func(field_type: FieldType, decode_binary=True):
    """Get convert func."""
    if (field_type == FieldType.C_VARCHAR or field_type == FieldType.C_BINARY) and not decode_binary:
        return _crow_varbinary_to_python
    return CONVERT_FUNC[field_type]


def convert_block_func(field_type: FieldType, decode_binary=True):
    """Get convert block func."""
    if (field_type == FieldType.C_VARCHAR or field_type == FieldType.C_BINARY) and not decode_binary:
        return _crow_varbinary_to_python_block
    return CONVERT_FUNC_BLOCK[field_type]


CONVERT_FUNC = {
    FieldType.C_BOOL: _crow_bool_to_python,
    FieldType.C_TINYINT: _crow_tinyint_to_python,
    FieldType.C_SMALLINT: _crow_smallint_to_python,
    FieldType.C_INT: _crow_int_to_python,
    FieldType.C_BIGINT: _crow_bigint_to_python,
    FieldType.C_FLOAT: _crow_float_to_python,
    FieldType.C_DOUBLE: _crow_double_to_python,
    FieldType.C_BINARY: _crow_binary_to_python,
    FieldType.C_TIMESTAMP: _crow_timestamp_to_python,
    FieldType.C_NCHAR: _crow_nchar_to_python,
    FieldType.C_TINYINT_UNSIGNED: _crow_tinyint_unsigned_to_python,
    FieldType.C_SMALLINT_UNSIGNED: _crow_smallint_unsigned_to_python,
    FieldType.C_INT_UNSIGNED: _crow_int_unsigned_to_python,
    FieldType.C_BIGINT_UNSIGNED: _crow_bigint_unsigned_to_python,
    FieldType.C_JSON: _crow_nchar_to_python,
    FieldType.C_VARBINARY: _crow_varbinary_to_python,
    FieldType.C_GEOMETRY: _crow_varbinary_to_python,
    FieldType.C_DECIMAL: _crow_decimal_to_python,
    FieldType.C_DECIMAL64: _crow_decimal_to_python,
}

CONVERT_FUNC_BLOCK = {
    FieldType.C_BOOL: _crow_bool_to_python,
    FieldType.C_TINYINT: _crow_tinyint_to_python,
    FieldType.C_SMALLINT: _crow_smallint_to_python,
    FieldType.C_INT: _crow_int_to_python,
    FieldType.C_BIGINT: _crow_bigint_to_python,
    FieldType.C_FLOAT: _crow_float_to_python,
    FieldType.C_DOUBLE: _crow_double_to_python,
    FieldType.C_BINARY: _crow_binary_to_python_block,
    FieldType.C_TIMESTAMP: _crow_timestamp_to_python,
    FieldType.C_NCHAR: _crow_nchar_to_python_block,
    FieldType.C_TINYINT_UNSIGNED: _crow_tinyint_unsigned_to_python,
    FieldType.C_SMALLINT_UNSIGNED: _crow_smallint_unsigned_to_python,
    FieldType.C_INT_UNSIGNED: _crow_int_unsigned_to_python,
    FieldType.C_BIGINT_UNSIGNED: _crow_bigint_unsigned_to_python,
    FieldType.C_JSON: _crow_nchar_to_python_block,
    FieldType.C_VARBINARY: _crow_varbinary_to_python_block,
    FieldType.C_GEOMETRY: _crow_varbinary_to_python_block,
    FieldType.C_DECIMAL: _crow_decimal_to_python,
    FieldType.C_DECIMAL64: _crow_decimal_to_python,
}


# Corresponding TAOS_FIELD structure in C


class TaosField(ctypes.Structure):
    _fields_ = [
        ("_name", ctypes.c_char * 65),
        ("_type", ctypes.c_uint8),
        ("_bytes", ctypes.c_uint16),
    ]

    @property
    def name(self):
        return self._name.decode("utf-8")

    @property
    def length(self):
        """alias to self.bytes"""
        return self._bytes

    @property
    def bytes(self):
        return self._bytes

    @property
    def type(self):
        return self._type

    def __dict__(self):
        return {"name": self.name, "type": self.type, "bytes": self.length}

    def __str__(self):
        return "{name: %s, type: %d, bytes: %d}" % (
            self.name,
            self.type,
            self.length,
        )

    def __getitem__(self, item):
        return getattr(self, item)


class TaosFields(object):
    def __init__(self, fields, count):
        if isinstance(fields, c_void_p):
            self._fields = cast(fields, POINTER(TaosField))
        if isinstance(fields, POINTER(TaosField)):
            self._fields = fields
        self._count = count
        self._iter = 0

    def as_ptr(self):
        return self._fields

    @property
    def count(self):
        return self._count

    @property
    def fields(self):
        return self._fields

    def __next__(self):
        return self._next_field()

    def next(self):
        return self._next_field()

    def _next_field(self):
        if self._iter < self.count:
            field = self._fields[self._iter]
            self._iter += 1
            return field
        else:
            raise StopIteration

    def __getitem__(self, item):
        return self._fields[item]

    def __iter__(self):
        self._iter = 0
        return self

    def __len__(self):
        return self.count



# encoding:UTF-8

from ctypes import *
from datetime import datetime, timezone

from taos.constants import FieldType
from taos.precision import PrecisionEnum, PrecisionError


from taos.field import get_tz

def parse_datetime_to_utc_timestamp(dt) -> int:
    if isinstance(dt, str):
        dt_naive = datetime.fromisoformat(dt)
    elif isinstance(dt, datetime):
        dt_naive = dt
    else:
        raise TypeError(f"dt type error, expected str or datetime type but got {type(dt)}.")
    #
    if dt_naive.tzinfo is None or dt_naive.tzinfo.utcoffset(dt_naive) is None:
        tz = get_tz()
        if tz is not None:
            dt_aware = dt_naive.replace(tzinfo=tz)
        else:
            dt_aware = dt_naive
    else:
        dt_aware = dt_naive.astimezone(timezone.utc)
    #
    result = int(dt_aware.timestamp())
    return result


def datetime_to_timestamp(value, precision, is_null_type=0):
    # type: (datetime | float | int | str | c_int64, PrecisionEnum, int) -> c_int64
    if is_null_type > 0 or value is None:
        return FieldType.C_BIGINT_NULL

    if type(value) is datetime or isinstance(value, str):
        utc_timestamp = parse_datetime_to_utc_timestamp(value)
        if precision == PrecisionEnum.Milliseconds:
            ret_value = c_int64(utc_timestamp * 10**3)
            return ret_value
        elif precision == PrecisionEnum.Microseconds:
            ret_value = c_int64(utc_timestamp * 10**6)
            return ret_value
        else:
            raise PrecisionError("datetime do not support nanosecond precision")
    elif type(value) is float:
        if precision == PrecisionEnum.Milliseconds:
            return c_int64(round(value * 10**3))
        elif precision == PrecisionEnum.Microseconds:
            return c_int64(round(value * 10**6))
        else:
            raise PrecisionError("time float do not support nanosecond precision")
    elif isinstance(value, int) and not isinstance(value, bool):
        return c_int64(value)
    elif isinstance(value, c_int64):
        return value
    return FieldType.C_BIGINT_NULL



from .errors import *
from .restclient import RestClient
from typing import Optional


class TaosRestCursor:
    """
    Implement [PEP 249 cursor API](https://peps.python.org/pep-0249/#cursor-objects)
    """

    def __init__(self, client: RestClient):
        self._c = client
        self.arraysize = 1
        self._rowcount = -1
        self._response = None
        self._index = -1
        self._description = None
        self._logfile = ""
        self._affected_rows = None

    @property
    def rowcount(self):
        """ the number of rows that the last .execute*() produced (for DQL statements like SELECT) or affected (for DML statements like UPDATE or INSERT)."""
        return self._rowcount

    @property
    def affected_rows(self):
        """Return the rowcount of insertion. For SELECT statement, it will be None"""
        return self._affected_rows

    @property
    def description(self):
        """
        Returns
        -------
        This read-only attribute is a sequence of 3-item sequences.
        Each of these sequences contains information describing one column:
        - column_name
        - type_name
        - internal_size
        """
        if self._response is None:
            return None
        return self._description

    def callproc(self, procname, parameters=None):
        raise NotSupportedError()

    def close(self):
        pass

    def execute(self, operation: str, parameters=None, req_id: Optional[int] = None):
        self._response = None
        self._index = -1
        self._response = self._c.sql(operation, req_id=req_id)

        if self._logfile:
            with open(self._logfile, "a", encoding="utf-8") as logfile:
                logfile.write(f"{operation};\n")

        if self._response["column_meta"][0][0] == 'affected_rows':
            # for INSERT
            self._description = self._response["column_meta"]
            self._affected_rows = self._response["data"][0][0]
            self._rowcount = self._affected_rows
            return self._affected_rows
        else:
            # for SELECT, Show, ...
            self._description = self._response["column_meta"]
            self._affected_rows = None
            self._rowcount = self._response["rows"]

    def log(self, logfile):
        self._logfile = logfile

    def executemany(self, operation: str, parameters=None, req_id: Optional[int] = None):
        self.execute(operation, parameters, req_id)

    def istype(self, col, datatype):
        if datatype.upper().strip() == self._description[col][1]:
            return True
        if datatype.upper().strip() == "BINARY" and self._description[col][1] == "VARCHAR":
            return True
        return False

    def get_type(self, col):
        if not self._description:
            return None
        return self._description[col][1]

    def fetchone(self):
        if self._response is None:
            raise OperationalError("no result to fetch")
        self._index += 1
        if self._index + 1 > self._response["rows"]:
            return None
        return self._response["data"][self._index]

    def fetchmany(self):
        return self.fetchone()

    def fetchall(self):
        if self._response is None:
            raise OperationalError("no result to fetch")
        start_index = self._index + 1
        self._index = self._response["rows"]
        return self._response["data"][start_index:]

    def nextset(self):
        raise NotSupportedError()

    def setinputsizes(self):
        raise NotSupportedError()

    def setoutputsize(self, size, column=None):
        raise NotSupportedError()

    def setoutputsize(self, size, column=None):
        raise NotSupportedError()



from .connection import TaosRestConnection, Result
from .cursor import TaosRestCursor
from .restclient import RestClient
from .errors import *


def connect(**kwargs) -> TaosRestConnection:
    """
   Keyword Arguments
   ----------------------------
   - url: str, optional, default "http://localhost:6041"
        url to connect
   - token: str, optional, default None
        TDengine cloud Token, which is required only by TDengine cloud service
   - user : str, optional, default root
       username used to log in
   - password : str, optional, default taosdata
       password used to log in
   - database : str, optional, default None
        default database to use.
   - timeout : int, optional.
       the optional timeout parameter specifies a timeout in seconds for blocking operations
   - convert_timestamp: bool, optional, default true
        whether to convert timestamp in RFC3339 format to python datatime.
   - timezone: str | datetime.tzinfo, optional, default None.
        When convert_timestamp is true, which timezone to used.
        When the type of timezone is str, it should be recognized by [pytz package](https://pypi.org/project/pytz/).
        When the timezone is None, system timezone will be used and the returned datetime object will be offset-naive (no tzinfo), otherwise the returned datetime will be offset-aware(with tzinfo)
   Examples
   -----------------------------
    connect to cloud service
    ```python
    import taosrest, os
    url = os.environ("TDENGINE_CLOUD_URL")
    token = os.environ("TDENGINE_ClOUD_TOKEN")
    conn = taosrest.connect(url=url, token=token)
    ```
   connect to local taosAdapter
   ```python
   import taosrest
   conn = taosrest.connect(url="http://localhost:6041")
   ```
   """
    return TaosRestConnection(**kwargs)



# encoding:UTF-8
from taos.cinterface import *
from taos.error import StatementError
from taos.result import TaosResult
from taos import bind2
from taos import log
from taos import utils
from taos.precision import PrecisionEnum, PrecisionError
from typing import Optional



_taos_async_fn_t = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int)

#
# bind data with table
#
class BindTable(object):
    def __init__(self, name, tags):
        self.name  = name
        self.tags  = tags
        self.datas = []

    # add column data
    def add_col_data(self, data):
        self.datas.append(data)


class TaosStmt2OptionImpl(ctypes.Structure):
    _fields_ = [
        ("reqid", ctypes.c_int64),
        ("singleStbInsert", ctypes.c_bool),
        ("singleTableBindOnce", ctypes.c_bool),
        ("asyncExecFn", _taos_async_fn_t),
        ("userdata", ctypes.c_void_p)
    ]


def taos_stmt2_async_exec(userdata, result_set, error_code):
    print(f"Executing asynchronously with userdata: {userdata}, result_set: {result_set}, error_code: {error_code}")


class TaosStmt2Option:
    def __init__(self, reqid: int=None, single_stb_insert: bool=False, single_table_bind_once: bool=False, **kwargs):
        self._impl = TaosStmt2OptionImpl()
        if reqid is None:
            reqid = utils.gen_req_id()
        else:
            if type(reqid) is not int:
                raise StatementError(f"reqid type error, expected int type but got {type(reqid)}.")

        self.reqid = reqid
        self.single_stb_insert = single_stb_insert
        self.single_table_bind_once = single_table_bind_once
        self.is_async = kwargs.get('is_async', False)
        self.async_exec_fn = taos_stmt2_async_exec if self.is_async else _taos_async_fn_t()
        self.userdata = ctypes.c_void_p(None)

    @property
    def reqid(self) -> int:
        return self._impl.reqid

    @reqid.setter
    def reqid(self, value: int):
        self._impl.reqid = ctypes.c_int64(value)

    @property
    def single_stb_insert(self) -> bool:
        return self._impl.singleStbInsert

    @single_stb_insert.setter
    def single_stb_insert(self, value: bool):
        self._impl.singleStbInsert = ctypes.c_bool(value)

    @property
    def single_table_bind_once(self) -> bool:
        return self._impl.singleTableBindOnce

    @single_table_bind_once.setter
    def single_table_bind_once(self, value: bool):
        self._impl.singleTableBindOnce = ctypes.c_bool(value)

    @property
    def async_exec_fn(self) -> _taos_async_fn_t:
        return self._impl.asyncExecFn

    @async_exec_fn.setter
    def async_exec_fn(self, value: _taos_async_fn_t):
        if value is not None and not isinstance(value, _taos_async_fn_t):
            raise TypeError("async_exec_fn must be an instance of _taos_async_fn_t or None")
        self._impl.asyncExecFn = value

    @property
    def userdata(self) -> ctypes.c_void_p:
        return self._impl.userdata

    @userdata.setter
    def userdata(self, value: ctypes.c_void_p):
        self._impl.userdata = value

    def get_impl(self):
        return self._impl

#
#    ------------- global fun define -------------
#

def checkConsistent(tbnames, tags, datas):
    # todo
    return True

def obtainSchema(statement2):
    if statement2._stmt2 is None:
        raise StatementError("stmt2 object is null.")

    try:
        count, fields = statement2.get_fields()
        statement2.tag_fields = []
        statement2.fields     = []
        for field in fields:
            if field.field_type == TAOS_FIELD_TAG:
                statement2.tag_fields.append(field)
            elif field.field_type == TAOS_FIELD_COL:
                statement2.fields.append(field)
        log.debug(f"obtain schema tag fields = {statement2.tag_fields}")
        log.debug(f"obtain schema fields     = {statement2.fields}")
    except Exception as err:
        log.debug(f"obtain schema tag/col fields failed, reason: {repr(err)}")
        return False

    return len(statement2.fields) > 0


def getField(statement2, index, isTag):
    if isTag:
        return statement2.tag_fields[index]
    else:
        return statement2.fields[index]

# create stmt2Bind list from tags
def createTagsBind(statement2, tagsTbs):
    binds = []
    # tables tags
    for tagsTb in tagsTbs:
        # table
        n = len(tagsTb)
        bindsTb = bind2.new_stmt2_binds(n)
        for i in range(n):
            field = getField(statement2, i, True)
            values = [tagsTb[i]]
            log.debug(f"tag i = {i} type={field.type} precision={field.precision} length={field.bytes}  values = {values}  tagsTb = {tagsTb}\n")
            bindsTb[i].set_value(field.type, values, field.precision)
        binds.append(bindsTb)

    return binds    

# create stmt2Bind list from columns
def createColsBind(statement2, colsTbs):
    binds = []
    # tables columns data
    for colsTb in colsTbs:
        # table
        n = len(colsTb)
        bindsTb = bind2.new_stmt2_binds(n)
        for i in range(n):
            field = getField(statement2, i, isTag=False)
            bindsTb[i].set_value(field.type, colsTb[i], field.precision)
        binds.append(bindsTb)

    return binds    


#
# create bindv from list
#
def createBindV(statement2, tbnames, tags, datas):

    if tbnames == None and tags == None and datas == None:
        raise StatementError("all bind params is None.")

    # count
    count  = -1
    ret = utils.detectListNone(tbnames)
    if ret   == utils.ALL_NONE:
        bindNames = None
    elif ret == utils.HAVE_NONE:
        raise StatementError("params tbnames some is None, some is not None, this is error.")
    else:
        # not found none
        if type(tbnames) not in [list, tuple]:
            raise StatementError(f"tbnames type error, expected list or tuple type but got {type(tbnames)}.")

        bindNames = tbnames 
        count = len(tbnames)

    # tags
    ret = utils.detectListNone(tags)
    if ret   == utils.ALL_NONE:
        bindTags = None
    elif ret == utils.HAVE_NONE:
        raise StatementError("params tags some is None, some is not None, this is error.")
    else:
        # not found none
        bindTags = createTagsBind(statement2, tags)
        if count == -1:
            count = len(tags)
        else:
            if count != len(tags):
                err = f"tags count is inconsistent. require count={count} real={len(tags)}"
                raise StatementError(err)

    # datas
    ret = utils.detectListNone(datas)
    if ret   == utils.ALL_NONE:
        bindDatas = None
    elif ret == utils.HAVE_NONE:
        raise StatementError("params datas some is None, some is not None, this is error.")
    else:
        # not found none
        bindDatas = createColsBind(statement2, datas)
        if count == -1:
            count = len(datas)
        else:
            if count != len(datas):
                err = f"datas count is inconsistent. require count={count} real={len(datas)}"
                raise StatementError(err)

    # create
    return bind2.new_bindv(count, bindNames, bindTags, bindDatas)




#
# -------------- stmt2 object --------------
#

class TaosStmt2(object):
    """TDengine STMT2 interface"""

    def __init__(self, stmt2, decode_binary=True):
        self._stmt2 = stmt2
        self._decode_binary = decode_binary
        self.fields     = None
        self.tag_fields = None
        self.types      = None
        self._is_insert = None

    def prepare(self, sql):
        if self._stmt2 is None:
            raise StatementError(ErrMsg.STMT2_NULL)

        if sql is None:
            raise StatementError("sql is null.")

        if not isinstance(sql, str):
            raise StatementError("sql is not str type.")

        if len(sql) == 0:
            raise StatementError("sql is empty.")

        self.fields     = None
        self.tag_fields = None
        self._is_insert = None

        taos_stmt2_prepare(self._stmt2, sql)

        # obtain schema if insert
        if self.is_insert():
            if obtainSchema(self) is False:
                raise StatementError(f"obtain schema failed, maybe this sql is not supported, sql: {sql}")

    def bind_param(self, tbnames, tags, datas):
        if self._stmt2 is None:
            raise StatementError(ErrMsg.STMT2_NULL)
        
        log.debug(f"bind_param tbnames  = {tbnames} \n")
        log.debug(f"bind_param tags     = {tags}    \n")
        log.debug(f"bind_param datasTbs = {datas}   \n")

        # check consistent
        if checkConsistent(tbnames, tags, datas) == False:
            raise StatementError("check consistent failed.")

        # bindV
        bindv = createBindV(self, tbnames, tags, datas)
        if bindv == None:
            raise StatementError("create stmt2 bindV failed.")
                
        # call engine
        taos_stmt2_bind_param(self._stmt2, bindv.get_address(), -1)

    # with table bind
    def bind_param_with_tables(self, tables):

        tbnames = []
        tagsTbs = []
        datasTbs = []

        for table in tables:
            tbnames.append(table.name)
            tagsTbs.append(table.tags)
            datasTbs.append(table.datas)

        # call bind
        self.bind_param(tbnames, tagsTbs, datasTbs)


    def execute(self) -> int:
        if self._stmt2 is None:
            raise StatementError("stmt2 object is null.")
        #
        self._affected_rows = taos_stmt2_exec(self._stmt2)
        return self._affected_rows

    def result(self):
        if self._stmt2 is None:
            raise StatementError("stmt2 object is null.")

        result = taos_stmt2_result(self._stmt2)
        return TaosResult(result, close_after=False, decode_binary=self._decode_binary)

    def close(self):
        if self._stmt2 is None:
            return 
        
        taos_stmt2_close(self._stmt2)
        self._stmt2 = None

    def is_insert(self):
        if self._stmt2 is None:
            raise StatementError("stmt2 object is null.")

        if self._is_insert is None:
            self._is_insert = taos_stmt2_is_insert(self._stmt2)

        return self._is_insert

    def get_fields(self):
        if self._stmt2 is None:
            raise StatementError("stmt2 object is null.")

        return taos_stmt2_get_fields(self._stmt2)

    def error(self):
        if self._stmt2 is None:
            raise StatementError("stmt2 object is null.")
        
        return taos_stmt2_error(self._stmt2)
    
    #
    # if query must set columns type (not engine interface),
    #  type is define class FieldType in constants.py
    #
    def set_columns_type(self, types):
        self.fields = []
        for type in types:
            item = TaosFieldAllCls(None, type, PrecisionEnum.Milliseconds, None, None, TAOS_FIELD_COL)
            self.fields.append(item)

    def __del__(self):
        self.close()

    @property
    def affected_rows(self):
        # type: () -> int
        return self._affected_rows



# encoding:UTF-8
import sys
import ctypes
from ctypes import *
from datetime import datetime, timezone
from typing import List, Optional

from taos.cinterface import IS_V3
from taos.constants import FieldType
from taos.precision import PrecisionEnum, PrecisionError
from taos import log
from taos import utils
from taos.field import get_tz
from taos.bind_base import datetime_to_timestamp


IS_NULL_TYPE_FALSE  = 0
IS_NULL_TYPE_TRUE   = 1
IS_NULL_TYPE_IGNORE = 2


class IgnoreUpdateType:
    def __init__(self, value=None):
        self.value = value

    def __repr__(self):
        return f"IgnoreUpdateType({self.value})"


IGNORE = IgnoreUpdateType()

def get_is_null_type(value) -> int:
    if value == IGNORE:
        return IS_NULL_TYPE_IGNORE
    elif value is None:
        return IS_NULL_TYPE_TRUE
    else:
        return IS_NULL_TYPE_FALSE


class TaosStmt2Bind(ctypes.Structure):
    _fields_ = [
        ("buffer_type", ctypes.c_int),
        ("buffer", ctypes.c_void_p),
        ("length", ctypes.POINTER(ctypes.c_int32)),
        ("is_null", ctypes.c_char_p),
        ("num", ctypes.c_int)
    ]


    #
    # set bind value with field type
    #
    def set_value(self, buffer_type, values, precision = PrecisionEnum.Milliseconds):
        log.debug(f"set_value type={buffer_type} precision={precision} values={values}\n")
        if values is not None:
            utils.checkTypeValid(buffer_type, values)
        if buffer_type == FieldType.C_BOOL:
            self.bool(values)
        elif buffer_type == FieldType.C_TINYINT:
            self.tinyint(values)
        elif buffer_type == FieldType.C_SMALLINT:
            self.smallint(values)
        elif buffer_type == FieldType.C_INT:
            self.int(values)
        elif buffer_type == FieldType.C_BIGINT:
            self.bigint(values)
        elif buffer_type == FieldType.C_FLOAT:
            self.float(values)
        elif buffer_type == FieldType.C_DOUBLE:
            self.double(values)
        elif buffer_type == FieldType.C_VARCHAR:
            self.varchar(values)
        elif buffer_type == FieldType.C_BINARY:
            self.binary(values)
        elif buffer_type == FieldType.C_TIMESTAMP:
            self.timestamp(values, precision)
        elif buffer_type == FieldType.C_NCHAR:
            self.nchar(values)
        elif buffer_type == FieldType.C_TINYINT_UNSIGNED:
            self.tinyint_unsigned(values)
        elif buffer_type == FieldType.C_SMALLINT_UNSIGNED:
            self.smallint_unsigned(values)
        elif buffer_type == FieldType.C_INT_UNSIGNED:
            self.int_unsigned(values)
        elif buffer_type == FieldType.C_BIGINT_UNSIGNED:
            self.bigint_unsigned(values)
        elif buffer_type == FieldType.C_JSON:
            self.json(values)
        elif buffer_type == FieldType.C_VARBINARY:
            self.varbinary(values)
        elif buffer_type == FieldType.C_GEOMETRY:
            self.geometry(values)

    def numeric_common(self, values, ctypes_type, buffer_null_type, buffer_value_type):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        #
        is_null = [get_is_null_type(value) for value in values]
        try:
            buffer = cast(values, c_void_p)
        except:
            buffer_type = ctypes_type * len(values)
            try:
                buffer = buffer_type(*values)
            except:
                buffer = buffer_type(*[buffer_null_type if is_null[idx] > 0 else value for idx, value in enumerate(values)])

        self.buffer = cast(buffer, c_void_p)
        self.buffer_type = buffer_value_type
        self.num = len(values)
        self.is_null = cast((c_char * len(values))(*is_null), c_char_p)

    def bool(self, values):
        cnt = len(values)
        for i in range(cnt):
            if type(values[i]) is float:
                values[i] = 0 if values[i] == 0 else 1

        #print(f"after values={values}")
        self.numeric_common(values, c_int8, FieldType.C_BOOL_NULL, FieldType.C_BOOL)

    def tinyint(self, values):
        self.numeric_common(values, c_int8, FieldType.C_TINYINT_NULL, FieldType.C_TINYINT)

    def smallint(self, values):
        self.numeric_common(values, c_int16, FieldType.C_SMALLINT_NULL, FieldType.C_SMALLINT)

    def int(self, values):
        self.numeric_common(values, c_int32, FieldType.C_INT_NULL, FieldType.C_INT)

    def bigint(self, values):
        self.numeric_common(values, c_int64, FieldType.C_BIGINT_NULL, FieldType.C_BIGINT)

    def float(self, values):
        self.numeric_common(values, c_float, FieldType.C_FLOAT_NULL, FieldType.C_FLOAT)

    def double(self, values):
        self.numeric_common(values, c_double, FieldType.C_DOUBLE_NULL, FieldType.C_DOUBLE)

    def timestamp(self, values, precision=PrecisionEnum.Milliseconds):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        #
        is_null = [get_is_null_type(value) for value in values]
        try:
            buffer = cast(values, c_void_p)
        except:
            buffer_type = c_int64 * len(values)
            buffer = buffer_type(*[datetime_to_timestamp(value, precision, is_null[idx]) for idx, value in enumerate(values)])

        self.buffer = cast(buffer, c_void_p)
        self.buffer_type = FieldType.C_TIMESTAMP
        self.num = len(values)
        self.is_null = cast((c_char * len(values))(*is_null), c_char_p)

    def _str_to_buffer(self, values, encode=True):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        #
        is_null = [get_is_null_type(value) for value in values]
        self.num = len(values)
        self.is_null = cast((c_byte * self.num)(*is_null), c_char_p)

        if sum([1 if v > 0 else 0 for v in is_null]) == self.num:
            self.length = (c_int32 * len(values))(0 * self.num)
            return

        _bytes_list = []
        if sys.version_info < (3, 0):
            _bytes_list = [None if is_null[idx] > 0 else bytes(value) for idx, value in enumerate(values)]
        else:
            if encode:
                _bytes_list = [None if is_null[idx] > 0 else value.encode("utf-8") for idx, value in enumerate(values)]
            else:
                _bytes_list = [None if is_null[idx] > 0 else bytes(value) for idx, value in enumerate(values)]
            #

        _bytes = b"".join([b for b in _bytes_list if b is not None])
        self.buffer = cast(create_string_buffer(_bytes), c_void_p)
        self.length = (c_int32 * len(values))(*[len(b) if b is not None else 0 for b in _bytes_list])

    def binary(self, values):
        self.buffer_type = FieldType.C_BINARY
        self._str_to_buffer(values)

    def nchar(self, values):
        # type: (list[str]) -> None
        self.buffer_type = FieldType.C_NCHAR
        self._str_to_buffer(values)

    def json(self, values):
        # type: (list[str]) -> None
        self.buffer_type = FieldType.C_JSON
        self._str_to_buffer(values)

    def tinyint_unsigned(self, values):
        self.numeric_common(values, c_uint8, FieldType.C_TINYINT_UNSIGNED_NULL, FieldType.C_TINYINT_UNSIGNED)

    def smallint_unsigned(self, values):
        self.numeric_common(values, c_uint16, FieldType.C_SMALLINT_UNSIGNED_NULL, FieldType.C_SMALLINT_UNSIGNED)

    def int_unsigned(self, values):
        self.numeric_common(values, c_uint32, FieldType.C_INT_UNSIGNED_NULL, FieldType.C_INT_UNSIGNED)

    def bigint_unsigned(self, values):
        self.numeric_common(values, c_uint64, FieldType.C_BIGINT_UNSIGNED_NULL, FieldType.C_BIGINT_UNSIGNED)

    def varchar(self, values):
        self.buffer_type = FieldType.C_VARCHAR
        self._str_to_buffer(values)

    def varbinary(self, values):
        self.buffer_type = FieldType.C_VARBINARY
        self._str_to_buffer(values, encode=False)

    def geometry(self, values):
        self.buffer_type = FieldType.C_GEOMETRY
        self._str_to_buffer(values, encode=False)


class TaosStmt2BindV(ctypes.Structure):
    _fields_ = [
        ("count", ctypes.c_int),
        ("tbnames", ctypes.POINTER(ctypes.c_char_p)),
        ("tags", ctypes.POINTER(ctypes.POINTER(TaosStmt2Bind))),
        ("bind_cols", ctypes.POINTER(ctypes.POINTER(TaosStmt2Bind)))
    ]

    def init(
            self,
            count: int,
            tbnames,    # List[str],
            tags,       # Optional[List[Array[TaosStmt2Bind]]],
            bind_cols   # Optional[List[Array[TaosStmt2Bind]]]
    ):
        self.count = count
        if tbnames is not None:
            self.tbnames = (ctypes.c_char_p * count)()
            for i, tbname in enumerate(tbnames):
                self.tbnames[i] = cast(self.str_to_buffer(tbname), c_char_p)
        else:
            self.tbnames = None

        if tags is not None:
            self.tags = (ctypes.POINTER(TaosStmt2Bind) * len(tags))()
            for i, tag_list in enumerate(tags):
                self.tags[i] = tag_list
        else:
            self.tags = None

        if bind_cols is not None:
            self.bind_cols = (ctypes.POINTER(TaosStmt2Bind) * count)()
            for i, col_list in enumerate(bind_cols):
                self.bind_cols[i] = col_list
        else:
            self.bind_cols = None
        #

    def str_to_buffer(self, value: str, encode=True):
        buffer = None
        if value is not None:
            _bytes = None
            if encode:
                _bytes = value.encode("utf-8")
            else:
                _bytes = bytes(value)

            buffer = cast(create_string_buffer(_bytes), c_void_p)
        else:
            buffer = None

        return buffer

    def get_address(self) -> ctypes.pointer:
        return c_void_p(ctypes.addressof(self))



def new_stmt2_binds(size: int) : # -> Array[TaosStmt2Bind]:
    return (TaosStmt2Bind * size)()


def new_bindv(
        count: int,
        tbnames,    # Optional[List[str]]
        tags,       # Optional[List[Array[TaosStmt2Bind]]],
        bind_cols   # Optional[List[Array[TaosStmt2Bind]]]
) -> TaosStmt2BindV:
    bindv = TaosStmt2BindV()
    bindv.init(count, tbnames, tags, bind_cols)
    return bindv


# encoding:UTF-8
import ctypes

from taos.constants import FieldType

_RTYPE = ctypes.c_uint16
_RTYPE_SIZE = ctypes.sizeof(_RTYPE)


def _crow_binary_to_python_block_v3(data, is_null, num_of_rows, offsets, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C binary row to python row."""
    assert offsets is not None
    res = []
    for i in range(abs(num_of_rows)):
        if offsets[i] == -1:
            res.append(None)
        else:
            rbyte = _RTYPE.from_address(data + offsets[i]).value
            chars = (ctypes.c_char * rbyte).from_address(data + offsets[i] + _RTYPE_SIZE).raw.decode("utf-8")
            res.append(chars)
    return res


def _crow_nchar_to_python_block_v3(data, is_null, num_of_rows, offsets, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C nchar row to python row."""
    assert offsets is not None
    res = []
    for i in range(abs(num_of_rows)):
        if offsets[i] == -1:
            res.append(None)
        else:
            rbyte = _RTYPE.from_address(data + offsets[i]).value
            chars = (ctypes.c_char * rbyte).from_address(data + offsets[i] + _RTYPE_SIZE).raw.decode("utf-8")
            res.append(chars)
    return res


def _crow_varbinary_to_python_block_v3(data, is_null, num_of_rows, offsets, precision=FieldType.C_TIMESTAMP_UNKNOWN):
    """Function to convert C varbinary row to python row."""
    assert offsets is not None
    res = []
    for i in range(abs(num_of_rows)):
        if offsets[i] == -1:
            res.append(None)
        else:
            rbyte = _RTYPE.from_address(data + offsets[i]).value
            chars = (ctypes.c_char * rbyte).from_address(data + offsets[i] + _RTYPE_SIZE).raw
            res.append(chars)
    return res


def convert_block_func_v3(field_type: FieldType, decode_binary=True):
    """Get convert block func."""
    if (field_type == FieldType.C_VARCHAR or field_type == FieldType.C_BINARY) and not decode_binary:
        return _crow_varbinary_to_python_block_v3
    return CONVERT_FUNC_BLOCK_v3[field_type]


CONVERT_FUNC_BLOCK_v3 = {
    FieldType.C_VARCHAR: _crow_binary_to_python_block_v3,
    FieldType.C_BINARY: _crow_binary_to_python_block_v3,
    FieldType.C_NCHAR: _crow_nchar_to_python_block_v3,
    FieldType.C_JSON: _crow_nchar_to_python_block_v3,
    FieldType.C_VARBINARY: _crow_varbinary_to_python_block_v3,
    FieldType.C_GEOMETRY: _crow_varbinary_to_python_block_v3,
}


# Corresponding TAOS_FIELD structure in C


class TaosField(ctypes.Structure):
    _fields_ = [
        ("_name", ctypes.c_char * 65),
        ("_type", ctypes.c_uint8),
        ("_bytes", ctypes.c_uint32),
    ]

    @property
    def name(self):
        return self._name.decode("utf-8")

    @property
    def length(self):
        """Alias to self.bytes."""
        return self._bytes

    @property
    def bytes(self):
        return self._bytes

    @property
    def type(self):
        return self._type

    def __dict__(self):
        """Construct dict."""
        return {"name": self.name, "type": self.type, "bytes": self.length}

    def __str__(self):
        """Construct str."""
        return "{name: %s, type: %d, bytes: %d}" % (self.name, self.type, self.length)

    def __getitem__(self, item):
        """Get attr."""
        return getattr(self, item)


class TaosFieldE(ctypes.Structure):
    _fields_ = [
        ("_name", ctypes.c_char * 65),
        ("_type", ctypes.c_int8),
        ("_precision", ctypes.c_uint8),
        ("_scale", ctypes.c_uint8),
        ("_bytes", ctypes.c_int32),
    ]

    @property
    def name(self):
        return self._name.decode("utf-8")

    @property
    def length(self):
        """Alias to self.bytes."""
        return self._bytes

    @property
    def bytes(self):
        return self._bytes

    @property
    def type(self):
        return self._type

    @property
    def precision(self):
        return self._precision

    @property
    def scale(self):
        return self._scale

    def __dict__(self):
        """Construct dict."""
        return {"name": self.name, "type": self.type, "precision": self.precision, "scale": self.scale, "bytes": self.length}

    def __str__(self):
        """Construct str."""
        return "{name: %s, type: %d, precision: %d, scale: %d, bytes: %d}" % (self.name, self.type, self.precision, self.scale, self.length)

    def __getitem__(self, item):
        """Get attr."""
        return getattr(self, item)


class TaosFields(object):
    def __init__(self, fields, count):
        """Init class."""
        if isinstance(fields, ctypes.c_void_p):
            self._fields = ctypes.cast(fields, ctypes.POINTER(TaosField))
        if isinstance(fields, ctypes.POINTER(TaosField)):
            self._fields = fields
        self._count = count
        self._iter = 0

    def as_ptr(self):
        """Return as ptr."""
        return self._fields

    @property
    def count(self):
        """Return count."""
        return self._count

    @property
    def fields(self):
        """Return fields."""
        return self._fields

    def __next__(self):
        """Next field."""
        return self._next_field()

    def next(self):
        """Next field."""
        return self._next_field()

    def _next_field(self):
        """Iter next_field."""
        if self._iter < self.count:
            field = self._fields[self._iter]
            self._iter += 1
        else:
            raise StopIteration
        return field

    def __getitem__(self, item):
        """Return field item."""
        return self._fields[item]

    def __iter__(self):
        """To iter."""
        self._iter = 0
        return self

    def __len__(self):
        """Get len."""
        return self.count

    def __str__(self):
        """Print"""
        return ",".join(str(f) for f in self)


class TaosFieldEs(object):
    def __init__(self, fields, count):
        """Init class."""
        if isinstance(fields, ctypes.c_void_p):
            self._fields = ctypes.cast(fields, ctypes.POINTER(TaosFieldE))
        if isinstance(fields, ctypes.POINTER(TaosField)):
            self._fields = fields
        self._count = count
        self._iter = 0

    def as_ptr(self):
        """Return as ptr."""
        return self._fields

    @property
    def count(self):
        """Return count."""
        return self._count

    @property
    def fields(self):
        """Return fields."""
        return self._fields

    def __next__(self):
        """Next field."""
        return self._next_field()

    def next(self):
        """Next field."""
        return self._next_field()

    def _next_field(self):
        """Iter next_field."""
        if self._iter < self.count:
            field = self._fields[self._iter]
            self._iter += 1
        else:
            raise StopIteration
        return field

    def __getitem__(self, item):
        """Return field item."""
        return self._fields[item]

    def __iter__(self):
        """To iter."""
        self._iter = 0
        return self

    def __len__(self):
        """Get len."""
        return self.count

    def __str__(self):
        """Print"""
        return ",".join(str(f) for f in self)


TAOS_FIELD_T = ctypes.c_int
class TaosFieldAll(ctypes.Structure):
    _fields_ = [
        ("_name", ctypes.c_char * 65),
        ("_type", ctypes.c_int8),
        ("_precision", ctypes.c_uint8),
        ("_scale", ctypes.c_uint8),
        ("_bytes", ctypes.c_int32),
        ("_field_type", ctypes.c_uint8)
    ]

    @property
    def name(self):
        return self._name.decode("utf-8")

    @property
    def type(self):
        return self._type

    @property
    def precision(self):
        return self._precision

    @property
    def scale(self):
        return self._scale

    @property
    def length(self):
        """Alias to self.bytes."""
        return self._bytes

    @property
    def bytes(self):
        return self._bytes

    @property
    def field_type(self):
        return self._field_type

    def __dict__(self):
        """Construct dict."""
        return {"name": self.name, "type": self.type, "precision": self.precision, "scale": self.scale, "bytes": self.length, "field_type": self.field_type}

    def __str__(self):
        """Construct str."""
        return "{name: %s, type: %d, precision: %d, scale: %d, bytes: %d field_type:%d}" % (self.name, self.type, self.precision, self.scale, self.length, self.field_type)

    def __getitem__(self, item):
        """Get attr."""
        return getattr(self, item)


class TaosFieldAllCls:
    def __init__(self, name, type, precision, scale, bytes_, field_type):
        self.name       = name
        self.type       = type
        self.precision  = precision
        self.scale      = scale
        self.bytes      = bytes_
        self.field_type = field_type

    def __repr__(self):
        return f"TaosFieldAllCls(name=\"{self.name}\", type={self.type}, precision={self.precision}, scale={self.scale}, bytes={self.bytes}, field_type={self.field_type})"




from taos.cinterface import *
from taos.error import StatementError
from taos.result import TaosResult


class TaosStmt(object):
    """TDengine STMT interface"""

    def __init__(self, stmt, decode_binary=True):
        self._stmt = stmt
        self._decode_binary = decode_binary

    def set_tbname(self, name):
        """Set table name if needed.

        Note that the set_tbname* method should only used in insert statement
        """
        if self._stmt is None:
            raise StatementError("Invalid use of set_tbname")
        taos_stmt_set_tbname(self._stmt, name)

    def prepare(self, sql):
        # type: (str) -> None
        taos_stmt_prepare(self._stmt, sql)

    def set_tbname_tags(self, name, tags):
        # type: (str, Array[TaosBind]) -> None
        """Set table name with tags, tags is array of BindParams"""
        if self._stmt is None:
            raise StatementError("Invalid use of set_tbname")
        taos_stmt_set_tbname_tags(self._stmt, name, tags)

    def bind_param(self, params, add_batch=True):
        # type: (Array[TaosBind], bool) -> None
        if self._stmt is None:
            raise StatementError("Invalid use of stmt")
        taos_stmt_bind_param(self._stmt, params)
        if add_batch:
            taos_stmt_add_batch(self._stmt)

    def bind_param_batch(self, binds, add_batch=True):
        # type: (Array[TaosMultiBind], bool) -> None
        if self._stmt is None:
            raise StatementError("Invalid use of stmt")
        taos_stmt_bind_param_batch(self._stmt, binds)
        if add_batch:
            taos_stmt_add_batch(self._stmt)

    def add_batch(self):
        if self._stmt is None:
            raise StatementError("Invalid use of stmt")
        taos_stmt_add_batch(self._stmt)

    def execute(self):
        if self._stmt is None:
            raise StatementError("Invalid use of execute")
        taos_stmt_execute(self._stmt)

    def use_result(self):
        """NOTE: Don't use a stmt result more than once."""
        result = taos_stmt_use_result(self._stmt)
        return TaosResult(result, close_after=False, decode_binary=self._decode_binary)

    @property
    def affected_rows(self):
        # type: () -> int
        return taos_stmt_affected_rows(self._stmt)

    def close(self):
        """Close stmt."""
        if self._stmt is None:
            return
        taos_stmt_close(self._stmt)
        self._stmt = None

    def __del__(self):
        self.close()
